import { DateComponent, memoize, mapHash, NowTimer } from '@fullcalendar/core/internal.js';
import { createElement } from '@fullcalendar/core/preact.js';
import { createDayHeaderFormatter, DayTableSlicer, DateHeaderCell, DayOfWeekHeaderCell } from '@fullcalendar/daygrid/internal.js';
import { ResourceDayTableJoiner, buildResourceHeaderTiers, ResourceHeaderCell } from '@fullcalendar/resource-daygrid/internal.js';
import { VResourceJoiner, flattenResources, VResourceSplitter, DEFAULT_RESOURCE_ORDER, ResourcelessDayTableModel, DayResourceTableModel, ResourceDayTableModel } from '@fullcalendar/resource/internal.js';
import { AllDaySplitter, buildDayRanges, splitSegsByCol, splitInteractionByCol, DayTimeColsSlicer, TimeGridLayout, TimeGridWeekNumber, buildTimeColsModel } from '@fullcalendar/timegrid/internal.js';

class ResourceDayTimeColsJoiner extends VResourceJoiner {
    transformSeg(seg, resourceDayTable, resourceI) {
        return [
            Object.assign(Object.assign({}, seg), { col: resourceDayTable.computeCol(seg.col, resourceI) }),
        ];
    }
}

class ResourceTimeGridView extends DateComponent {
    constructor() {
        super(...arguments);
        this.flattenResources = memoize(flattenResources);
        this.buildResourceTimeColsModel = memoize(buildResourceTimeColsModel);
        this.allDaySplitter = new AllDaySplitter();
        // for all-day-resource props
        this.allDayResourceSplitter = new VResourceSplitter();
        this.allDayResourceSlicers = {};
        this.allDayResourceJoiner = new ResourceDayTableJoiner();
        // for timed resource props
        this.buildDayRanges = memoize(buildDayRanges);
        this.timedResourceSplitter = new VResourceSplitter();
        this.timedResourceSlicers = {};
        this.timedResourceJoiner = new ResourceDayTimeColsJoiner();
        // timed-only column splitting
        this.splitFgEventSegs = memoize(splitSegsByCol);
        this.splitBgEventSegs = memoize(splitSegsByCol);
        this.splitBusinessHourSegs = memoize(splitSegsByCol);
        this.splitNowIndicatorSegs = memoize(splitSegsByCol);
        this.splitDateSelectionSegs = memoize(splitSegsByCol);
        this.splitEventDrag = memoize(splitInteractionByCol);
        this.splitEventResize = memoize(splitInteractionByCol);
        // other memo
        this.createDayHeaderFormatter = memoize(createDayHeaderFormatter);
        this.isHitComboAllowed = (hit0, hit1) => {
            let allowAcrossResources = this.dayRanges.length === 1;
            return allowAcrossResources || hit0.dateSpan.resourceId === hit1.dateSpan.resourceId;
        };
    }
    render() {
        let { props, context } = this;
        let { options, dateEnv } = context;
        let { dateProfile } = props;
        let resourceOrderSpecs = options.resourceOrder || DEFAULT_RESOURCE_ORDER;
        let resources = this.flattenResources(props.resourceStore, resourceOrderSpecs);
        let resourceDayTableModel = this.buildResourceTimeColsModel(dateProfile, context.dateProfileGenerator, resources, options.datesAboveResources, context);
        // split seg by all-day/timed
        let splitProps = this.allDaySplitter.splitProps(props);
        // split the all-day segs by resource
        let allDayResourceSplitProps = this.allDayResourceSplitter.splitProps({
            businessHours: splitProps.allDay.businessHours,
            dateSelection: splitProps.allDay.dateSelection,
            eventStore: splitProps.allDay.eventStore,
            eventUiBases: splitProps.allDay.eventUiBases,
            eventSelection: splitProps.allDay.eventSelection,
            eventDrag: splitProps.allDay.eventDrag,
            eventResize: splitProps.allDay.eventResize,
            resourceDayTableModel,
        });
        this.allDayResourceSlicers = mapHash(allDayResourceSplitProps, (split, resourceId) => this.allDayResourceSlicers[resourceId] || new DayTableSlicer());
        let allDayResourceSlicedProps = mapHash(this.allDayResourceSlicers, (slicer, resourceId) => slicer.sliceProps(allDayResourceSplitProps[resourceId], dateProfile, options.nextDayThreshold, context, resourceDayTableModel.dayTableModel));
        let allDayResourceJoinedProps = this.allDayResourceJoiner.joinProps(allDayResourceSlicedProps, resourceDayTableModel);
        // split the timed segs by resource
        let dayRanges = this.dayRanges = this.buildDayRanges(resourceDayTableModel.dayTableModel, dateProfile, dateEnv);
        let timedResourceSplitProps = this.timedResourceSplitter.splitProps({
            businessHours: splitProps.timed.businessHours,
            dateSelection: splitProps.timed.dateSelection,
            eventStore: splitProps.timed.eventStore,
            eventUiBases: splitProps.timed.eventUiBases,
            eventSelection: splitProps.timed.eventSelection,
            eventDrag: splitProps.timed.eventDrag,
            eventResize: splitProps.timed.eventResize,
            resourceDayTableModel,
        });
        this.timedResourceSlicers = mapHash(timedResourceSplitProps, (split, resourceId) => this.timedResourceSlicers[resourceId] || new DayTimeColsSlicer());
        let timedResourceSlicedProps = mapHash(this.timedResourceSlicers, (slicer, resourceId) => slicer.sliceProps(timedResourceSplitProps[resourceId], dateProfile, null, context, dayRanges));
        let timedResourceJoinedProps = this.timedResourceJoiner.joinProps(timedResourceSlicedProps, resourceDayTableModel);
        let datesRepDistinctDays = resourceDayTableModel.dayTableModel.rowCnt === 1;
        let headerTiers = buildResourceHeaderTiers(// TODO: memoize
        resources, resourceDayTableModel.dayTableModel.headerDates, options.datesAboveResources, datesRepDistinctDays, context);
        let dayHeaderFormat = this.createDayHeaderFormatter(context.options.dayHeaderFormat, datesRepDistinctDays, resourceDayTableModel.colCnt);
        return (createElement(NowTimer, { unit: options.nowIndicator ? 'minute' : 'day' /* hacky */ }, (nowDate, todayRange) => {
            // timed-only column splitting
            let colCnt = resourceDayTableModel.colCnt;
            let fgEventSegsByCol = this.splitFgEventSegs(timedResourceJoinedProps.fgEventSegs, colCnt);
            let bgEventSegsByCol = this.splitBgEventSegs(timedResourceJoinedProps.bgEventSegs, colCnt);
            let businessHourSegsByCol = this.splitBusinessHourSegs(timedResourceJoinedProps.businessHourSegs, colCnt);
            let nowIndicatorSegsByCol = this.splitNowIndicatorSegs((() => {
                // was buildNowIndicatorSegs
                let nonResourceSegs = this.timedResourceSlicers[''].sliceNowDate(nowDate, this.props.dateProfile, this.context.options.nextDayThreshold, this.context, this.dayRanges);
                return this.timedResourceJoiner.expandSegs(resourceDayTableModel, nonResourceSegs);
            })(), colCnt);
            let dateSelectionSegsByCol = this.splitDateSelectionSegs(timedResourceJoinedProps.dateSelectionSegs, colCnt);
            let eventDragByCol = this.splitEventDrag(timedResourceJoinedProps.eventDrag, colCnt);
            let eventResizeByCol = this.splitEventResize(timedResourceJoinedProps.eventResize, colCnt);
            return (createElement(TimeGridLayout, { dateProfile: dateProfile, nowDate: nowDate, todayRange: todayRange, cells: resourceDayTableModel.cells[0], forPrint: props.forPrint, isHitComboAllowed: this.isHitComboAllowed, className: 'fc-resource-timegrid-view', 
                // header content
                headerTiers: headerTiers, renderHeaderLabel: (tierNum, innerWidthRef, innerHeightRef, width) => ((options.weekNumbers && tierNum === headerTiers.length - 1) ? (createElement(TimeGridWeekNumber // .fc-cell
                , { dateProfile: dateProfile, innerWidthRef: innerWidthRef, innerHeightRef: innerHeightRef, width: width })) : (createElement("div", { className: 'fc-cell', style: { width } }))), 
                // TODO: DRY
                renderHeaderContent: (model, tierNum, innerHeightRef, colWidth) => {
                    if (model.resource) {
                        return (createElement(ResourceHeaderCell, Object.assign({}, model, { innerHeightRef: innerHeightRef, colSpan: model.colSpan, colWidth: colWidth, isSticky: tierNum < headerTiers.length - 1 })));
                    }
                    else if (model.date) {
                        return (createElement(DateHeaderCell, Object.assign({}, model, { navLink: resourceDayTableModel.dayTableModel.colCnt > 1, dateProfile: props.dateProfile, todayRange: todayRange, dayHeaderFormat: dayHeaderFormat, innerHeightRef: innerHeightRef, colSpan: model.colSpan, colWidth: colWidth })));
                    }
                    else {
                        createElement(DayOfWeekHeaderCell, Object.assign({}, model, { dayHeaderFormat: dayHeaderFormat, innerHeightRef: innerHeightRef, colSpan: model.colSpan, colWidth: colWidth }));
                    }
                }, getHeaderModelKey: (model) => (model.resource
                    ? model.resource.id
                    : model.date
                        ? model.date.toISOString()
                        : model.dow), 
                // all-day content
                fgEventSegs: allDayResourceJoinedProps.fgEventSegs, bgEventSegs: allDayResourceJoinedProps.bgEventSegs, businessHourSegs: allDayResourceJoinedProps.businessHourSegs, dateSelectionSegs: allDayResourceJoinedProps.dateSelectionSegs, eventDrag: allDayResourceJoinedProps.eventDrag, eventResize: allDayResourceJoinedProps.eventResize, 
                // timed content
                fgEventSegsByCol: fgEventSegsByCol, bgEventSegsByCol: bgEventSegsByCol, businessHourSegsByCol: businessHourSegsByCol, nowIndicatorSegsByCol: nowIndicatorSegsByCol, dateSelectionSegsByCol: dateSelectionSegsByCol, eventDragByCol: eventDragByCol, eventResizeByCol: eventResizeByCol, 
                // universal content
                eventSelection: allDayResourceJoinedProps.eventSelection }));
        }));
    }
}
/*
TODO: kill this and DayResourceTableModel/ResourceDayTableModel
*/
function buildResourceTimeColsModel(dateProfile, dateProfileGenerator, resources, datesAboveResources, context) {
    let dayTable = buildTimeColsModel(dateProfile, dateProfileGenerator);
    if (!resources.length) {
        return new ResourcelessDayTableModel(dayTable);
    }
    return datesAboveResources ?
        new DayResourceTableModel(dayTable, resources, context) :
        new ResourceDayTableModel(dayTable, resources, context);
}

export { ResourceTimeGridView };
